"""

Dataclass field helpers.
"""

from dataclasses import field, Field
from typing import TypeVar

T = TypeVar('T')


def injected(type_, *args, **kwargs) -> Field:
    """Get a value from an injected type

    We frequently want part of an injectable or adapter. Just an attribute
    or, for injectable/adapters that are dict-like, a key.

    This field type accepts an injectable/adapter and serves three purposes:

    - Get an attribute off of it by passing in attr='somevalue'

    - Or, get a key out of it by passing in key=''

    - Or, if the injectable/adapter has a __call__, return the value
      of calling it

    - Otherwise, return the injectable/adapter, same as if you didn't
      provide a field

     We could just manually do ``injected`` then pick apart the injected
     value. But that exposes a big surface area. This field type lets us
     zero in on what we want.
    """

    # If metadata was also passed in, preserve it, otherwise start clean
    if 'metadata' not in kwargs:
        kwargs['metadata'] = {}

    kwm = kwargs['metadata']
    kwm['injected'] = dict(type_=type_, pipeline=args)

    return field(**kwargs)



project = 'viewdom_wired'
html_title = 'viewdom_wired'
copyright = '2020, Paul Everitt <pauleveritt@me.com>'
author = 'Paul Everitt <pauleveritt@me.com>'
release = '0.0.1'
extensions = [
    'sphinx.ext.autodoc',
    'myst_parser',
]
templates_path = ['_templates']
html_theme = 'sphinx_book_theme'
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
html_static_path = ['_static']
html_theme_options = dict(
    extra_footer='Theme by the <a href="https://ebp.jupyterbook.org">'
    + 'Executable Book Project</a>.',
    repository_url='https://github.com/pauleveritt/viewdom_wired',
    use_repository_button=True,
)
html_css_files = [
    'custom.css',
]
html_sidebars = {
    "**": [
        'subtitle.html',
        'sidebar-search-bs.html',
        'sbt-sidebar-nav.html',
    ]
}
myst_admonition_enable = True



from viewdom import Context
from viewdom.h import flatten, escape, encode_prop, VDOMNode, VOIDS
from wired import ServiceContainer
from wired_injector.injector import Injector


def relaxed_call(
    injector: Injector,
    callable_,
    children=None,
    parent_component=None,
    **kwargs,
):
    """ Make a component instance then call its __call__, returning a VDOM """

    target = injector.container.get(callable_)
    system_props = dict(children=children, parent_component=parent_component)
    component = injector(target, system_props=system_props, **kwargs)
    return component


def render(value, container: ServiceContainer, **kwargs):
    injector = Injector(container)
    return "".join(
        render_gen(
            Context(value, **kwargs),
            injector=injector,
            children=None,
            parent_component=None,
        )
    )


def render_gen(
    value, injector: Injector, children=None, parent_component=None
):
    for item in flatten(value):
        if isinstance(item, VDOMNode):
            tag, props, children = item.tag, item.props, item.children
            if callable(tag):
                component = relaxed_call(
                    injector,
                    tag,
                    children=children,
                    parent_component=parent_component,
                    **props,
                )
                parent_component = component
                yield from render_gen(
                    component(), injector, children, parent_component
                )
                continue

            yield f"<{escape(tag)}"
            if props:
                pi = props.items()
                yield f" {' '.join(encode_prop(k, v) for (k, v) in pi)}"

            if children:
                yield ">"
                yield from render_gen(
                    children, injector, parent_component=parent_component
                )
                yield f'</{escape(tag)}>'
            elif tag.lower() in VOIDS:
                yield '/>'
            else:
                yield f'></{tag}>'
        elif item not in (True, False, None):
            yield escape(item)



import os
import sys

import pytest


@pytest.fixture(scope="session", autouse=True)
def examples_path():
    """ Automatically add the root of the repo to path """
    ep = os.path.abspath(
        os.path.join(os.path.dirname(__file__), '..')
    )
    sys.path.insert(0, ep)



from dataclasses import dataclass, field
from typing import List

from viewdom.h import html, VDOM
from wired import ServiceRegistry
from wired_injector.decorators import register_injectable

pytest_plugins = [
    'viewdom_wired.fixtures',
]


@dataclass
class Heading:
    name: str = 'Hello'

    def __call__(self) -> VDOM:
        return html('''<h1>{self.name}</h1>''')


@dataclass
class Person:
    first_name: str
    last_name: str
    full_name: List[str] = field(init=False)

    def __post_init__(self):
        self.full_name = [self.first_name, self.last_name]

    def __call__(self):
        full_name = ' '.join(self.full_name)  # noqa: F841
        return html('<div>{full_name}</div>')


def test_wired_renderer_simplest_nocontainer(registry: ServiceRegistry):
    from viewdom.h import render

    instance = Heading(name='No Wired')
    expected = '<h1>No Wired</h1>'
    actual = render(instance())
    assert expected == actual


def test_wired_renderer_simplest_container(registry: ServiceRegistry):
    from viewdom_wired import render

    container = registry.create_container()
    register_injectable(registry, Heading)
    expected = '<h1>Hello</h1>'
    actual = render(html('''<{Heading}/>'''), container)
    assert expected == actual


def test_wired_renderer_simplest_propoverride(registry: ServiceRegistry):
    from viewdom_wired import render

    container = registry.create_container()
    register_injectable(registry, Heading)
    expected = '<h1>Override</h1>'
    actual = render(html('''<{Heading} name="Override"/>'''), container)
    assert expected == actual


def test_wired_renderer_simplest_init_false(registry: ServiceRegistry):
    from viewdom_wired import render

    container = registry.create_container()
    register_injectable(registry, Person)
    expected = '<div>Paul Everitt</div>'
    actual = render(
        html(
            '''
    <{Person} first_name="Paul" last_name="Everitt"/>
    '''
        ),
        container,
    )
    assert expected == actual


def test_wired_renderer_non_void(registry: ServiceRegistry):
    @dataclass
    class NonVoid:
        def __call__(self):
            return html('<i class="icon"></i>')

    from viewdom_wired import render

    container = registry.create_container()
    register_injectable(registry, NonVoid)
    expected = '<i class="icon"></i>'
    actual = render(html('''<{NonVoid} />'''), container)
    assert expected == actual



"""
General-purpose components useful in all viewdom_wired apps.

"""
from dataclasses import dataclass

from viewdom import VDOM


@dataclass
class Super:
    parent_children: VDOM = ()

    def __call__(self) -> VDOM:
        return self.parent_children



from typing import TypeVar, Callable, Optional, Any, Type

from venusian import Scanner, attach
from wired import ServiceContainer, ServiceRegistry

protocol = TypeVar("protocol")


def adherent(
    c: Callable[[], protocol]
) -> Callable[[Type[protocol]], Type[protocol]]:
    def decor(input_value: Type[protocol]) -> Type[protocol]:
        return input_value

    return decor


def register_component(
    registry: ServiceRegistry,
    for_: Callable,
    target: Callable = None,
    context: Optional[Any] = None,
):
    """ Imperative form of the component decorator """

    def component_factory(container: ServiceContainer):
        return target if target else for_

    registry.register_factory(component_factory, for_, context=context)


class component:
    def __init__(self, for_: type = None, context: Type = None):
        self.for_ = for_
        self.context = context

    def __call__(self, wrapped):
        def callback(scanner: Scanner, name: str, cls):
            for_ = self.for_ if self.for_ else cls
            registry: ServiceRegistry = getattr(scanner, 'registry')

            register_component(
                registry,
                for_,
                target=cls,
                context=self.context,
            )

        attach(wrapped, callback, category='wired_component')
        return wrapped



from dataclasses import dataclass
from typing import Protocol

import pytest
from viewdom import VDOM
from wired import ServiceRegistry, ServiceContainer
from wired.dataclasses import factory, register_dataclass


class Logo(Protocol):
    src: str
    alt: str

    def __call__(self) -> VDOM:
        ...


@dataclass
class Customer:
    name: str = 'Some Customer'


@factory()
@dataclass
class Settings:
    greeting: str = 'Hello'


@pytest.fixture
def registry() -> ServiceRegistry:
    import sys
    from venusian import Scanner

    registry = ServiceRegistry()
    scanner = Scanner(registry=registry)
    current_module = sys.modules[__name__]
    scanner.scan(current_module)
    register_dataclass(registry, Settings)
    return registry


@pytest.fixture
def customer() -> Customer:
    return Customer()


@pytest.fixture
def container(registry, customer) -> ServiceContainer:
    c = registry.create_container(context=customer)
    return c



"""
The component renderer uses a variation of wired's dataclass dependency
injector. Write some tests that cover policies.

"""
from dataclasses import dataclass

import pytest
from viewdom import html, VDOM
from wired_injector.decorators import register_injectable
from wired_injector.injector import Injector
from wired_injector.operators import Context, Attr

from viewdom_wired.fixtures import Customer

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated  # type: ignore

pytest_plugins = [
    'viewdom_wired.fixtures',
]


class Person:
    """ A marker class """

    pass


@pytest.fixture
def injector(container) -> Injector:
    injector = Injector(container)
    return injector


def test_str_default_value(registry, injector):
    """ Simple type (str) and has a default """

    @dataclass
    class TestPerson:
        name: str = 'default'

        def __call__(self) -> VDOM:
            return html('<div>{self.name}</div>')

    register_injectable(registry, for_=Person, target=TestPerson)
    person: TestPerson = injector(TestPerson)
    assert 'default' == person.name


def test_str_prop(registry, injector):
    """ Simple type (str) with passed-in value """

    @dataclass
    class TestPerson:
        name: str = 'default'

        def __call__(self) -> VDOM:
            return html('<div>{self.name}</div>')

    register_injectable(registry, for_=Person, target=TestPerson)
    person: TestPerson = injector(TestPerson, name='passed in')
    assert 'passed in' == person.name


def test_context(registry, injector):
    """ Use the type-hint to inject the context """

    @dataclass
    class TestPerson:
        customer: Annotated[
            Customer,
            Context(),
        ]

        def __call__(self) -> VDOM:
            return html('<div>{self.customer.name}</div>')

    register_injectable(
        registry, for_=Person, target=TestPerson, context=Customer
    )
    person = injector(TestPerson)
    assert 'Some Customer' == person.customer.name


def test_injected_attr(registry, injector):
    """ Used ``injected`` to get the context and grab an attr """

    @dataclass
    class TestPerson:
        name: Annotated[
            str,
            Context(),
            Attr('name'),
        ]

        def __call__(self) -> VDOM:
            return html('<div>{self.name}</div>')

    register_injectable(registry, for_=Person, target=TestPerson)
    person = injector(TestPerson)
    assert 'Some Customer' == person.name



"""

Test usage of wired and components via decorators instead of
imperative.

More cumbersome (due to scanner) to copy around so placed into a
single test.

"""
from dataclasses import dataclass

import pytest
from viewdom.h import html
from wired import ServiceRegistry
from wired.dataclasses import factory, register_dataclass
from wired_injector import injectable
from wired_injector.operators import Attr, Context, Get

from viewdom_wired import render

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated  # type: ignore


class FirstContext:
    def __init__(self):
        self.name = 'First Context'


class SecondContext:
    def __init__(self):
        self.name = 'Second Context'


@factory()
@dataclass
class Settings:
    greeting: str = 'Hello'


@injectable()
@dataclass
class Heading:
    person: str
    name: Annotated[str, Context(), Attr('name')]
    greeting: Annotated[str, Get(Settings), Attr('greeting')]

    def __call__(self):
        return html('''<h1>{self.greeting} {self.person}, {self.name}</h1>''')


@pytest.fixture
def registry() -> ServiceRegistry:
    import sys
    from venusian import Scanner

    registry = ServiceRegistry()
    scanner = Scanner(registry=registry)
    current_module = sys.modules[__name__]
    scanner.scan(current_module)
    register_dataclass(registry, Settings)
    return registry


@injectable(for_=Heading, context=SecondContext)
@dataclass
class SecondHeading:
    person: str
    name: Annotated[str, Context(), Attr('name')]
    greeting: Annotated[str, Get(Settings), Attr('greeting')]

    def __call__(self):
        return html(
            '''
        <h1>{self.greeting} {self.person}... {self.name}</h1>
        '''
        )


def test_wired_renderer_first(registry: ServiceRegistry):
    container = registry.create_container(context=FirstContext())
    expected = '<h1>Hello World, First Context</h1>'
    actual = render(html('''<{Heading} person="World"/>'''), container)
    assert expected == actual


def test_wired_renderer_second(registry: ServiceRegistry):
    container = registry.create_container(context=SecondContext())
    expected = '<h1>Hello World... Second Context</h1>'
    actual = render(html('''<{Heading} person="World"/>'''), container)
    assert expected == actual



from dataclasses import dataclass
from typing import Tuple

import pytest
from viewdom.h import html
from wired import ServiceRegistry
from wired.dataclasses import factory, register_dataclass
from wired_injector import injectable
from wired_injector.decorators import register_injectable
from wired_injector.operators import Context, Attr, Get

from viewdom_wired import render

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated  # type: ignore


class FirstContext:
    def __init__(self):
        self.name = 'First Context'


class SecondContext:
    def __init__(self):
        self.name = 'Second Context'


@factory()
@dataclass
class Settings:
    greeting: str = 'Hello'


@injectable()
@dataclass
class Heading:
    person: str
    name: Annotated[
        str,
        Context(),
        Attr('name'),
    ]
    greeting: Annotated[
        str,
        Get(Settings),
        Attr('greeting'),
    ]

    def __call__(self):
        return html('''<h1>{self.greeting} {self.person}, {self.name}</h1>''')


@pytest.fixture
def registry() -> ServiceRegistry:
    import sys
    from venusian import Scanner

    registry = ServiceRegistry()
    scanner = Scanner(registry=registry)
    current_module = sys.modules[__name__]
    scanner.scan(current_module)
    register_dataclass(registry, Settings)
    return registry


@injectable(for_=Heading, context=SecondContext)
@dataclass
class SecondHeading:
    person: str
    name: Annotated[
        str,
        Context(),
        Attr('name'),
    ]
    greeting: Annotated[
        str,
        Get(Settings),
        Attr('greeting'),
    ]

    def __call__(self):
        return html(
            '''
        <h1>{self.greeting} {self.person}... {self.name}</h1>
        '''
        )


def test_wired_renderer_first(registry: ServiceRegistry):
    container = registry.create_container(context=FirstContext())
    expected = '<h1>Hello World, First Context</h1>'
    actual = render(html('''<{Heading} person="World"/>'''), container)
    assert expected == actual


def test_wired_renderer_second(registry: ServiceRegistry):
    container = registry.create_container(context=SecondContext())
    expected = '<h1>Hello World... Second Context</h1>'
    actual = render(html('''<{Heading} person="World"/>'''), container)
    assert expected == actual


def test_wired_renderer_children(registry: ServiceRegistry):
    @dataclass
    class Heading2:
        children: str
        name: str = 'Hello'

        def __call__(self):
            return html('''<h1>{self.name}</h1><div>{self.children}</div>''')

    registry = ServiceRegistry()
    register_injectable(registry, Heading, Heading2)
    container = registry.create_container()
    expected = '<h1>Hello</h1><div>Child</div>'
    actual = render(html('''<{Heading}>Child<//>'''), container)
    assert expected == actual


def test_wired_renderer_generics(registry: ServiceRegistry):
    """ Tolerate usage of PEP 484 generics on type hints """

    @dataclass
    class LocalHeading:
        names: Tuple[str, ...] = ('Name 1',)

        def __call__(self):
            name_one = self.names[0]  # noqa: F841
            return html('<h1>{name_one}</h1>')

    container = registry.create_container()
    register_injectable(registry, LocalHeading, LocalHeading)
    expected = '<h1>Name 1</h1>'
    actual = render(html('''<{LocalHeading}/>'''), container)
    assert expected == actual



import pytest

from examples import (
    hello_world,
    hello_children,
    hello_app,
    app_plugin_site,
    app_decorators_render,
    subcomponents,
    override,
    context,
    custom_context,
    protocols_hello_logo,
    logo_protocol,
    adherent,
)


@pytest.mark.parametrize(
    'target',
    [
        hello_world,
        hello_children,
        hello_app,
        app_plugin_site,
        app_decorators_render,
        subcomponents,
        override,
        context,
        custom_context,
        protocols_hello_logo,
        logo_protocol,
        adherent,
    ],
)
def test_examples(target):
    expected, actual = target.test()
    assert expected == actual



from setuptools import setup, find_packages


def readfile(name):
    with open(name) as f:
        return f.read()


readme = readfile('README.rst')
changes = readfile('CHANGES.rst')

requires = [
    'tagged',
    'htm',
    'viewdom',
    'MarkupSafe',
    'typing_extensions>=3.7.4;python_version<"3.8"',
    'wired',
    'wired_injector',
    'venusian',
]

docs_require = [
    'Sphinx',
    'sphinx-book-theme',
    'myst-parser',
]

tests_require = [
    'pytest',
]

dev_require = [
    'mypy',
    'coverage',
    'pytest-coverage',
    'tox',
    'black',
    'flake8',
    'twine',
    'pre-commit',
    'check-manifest',
]


setup(
    name='viewdom_wired',
    description=(
        'Injectable and plugabble components and services for viewdom'
    ),
    version='0.1.0',
    long_description=readme + '\n\n' + changes,
    long_description_content_type='text/x-rst',
    author='Paul Everitt',
    author_email='pauleveritt@me.com',
    url='https://viewdom_wired.readthedocs.io',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={"viewdom_wired": ["py.typed"]},
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=requires,
    extras_require=dict(
        docs=docs_require,
        tests=tests_require,
        dev=dev_require,
    ),
    zip_safe=False,
    keywords=','.join(
        [
            'web',
            'html',
            'components',
            'templates',
        ]
    ),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
)
